# Обозначения весов интервала
import copy

import numpy

UNISON = 0
MINOR_SECOND = 1
MAJOR_SECOND = .8
MINOR_THIRD = .225
MAJOR_THIRD = .2
PERFECT_FOURTH = .55
AUGMENTED_FOURTH = .65
PERFECT_FIFTH = .1
MINOR_SIXTH = .275
MAJOR_SIXTH = .25
MINOR_SEVENTH = .7
MAJOR_SEVENTH = .9
PERFECT_OCTAVE = 0

MINOR_NINTH = 1
MAJOR_NINTH = .8
MINOR_TENTH = .225
MAJOR_TENTH = .2
PERFECT_ELEVENTH = .55
AUGMENTED_ELEVENTH = .65
PERFECT_TWELVETH = .1
MINOR_THIRTEENTH = .275
MAJOR_THIRTEENTH = .25
MINOR_FOURTEENTH = .7
MAJOR_FOURTEENTH = .9
PERFECT_FIFTEENTH = 0
MINOR_SIXTEENTH = 1
MAJOR_SIXTEENTH = .8
MINOR_SEVENTEENTH = .225
MAJOR_SEVENTEENTH = .2
PERFECT_EIGHTEENTH = .55
AUGMENTED_EIGHTEENTH = .65
PERFECT_NINETEENTH = .1
MINOR_TWENTIETH = .275
MAJOR_TWENTIETH = .25
MINOR_TWENTYFIRST = .7
MAJOR_TWENTYFIRST = .9
PERFECT_TWENTYSECOND = 0
MINOR_TWENTYTHIRD = 1
MAJOR_TWENTYTHIRD = .8
MINOR_TWENTYFOURTH = .225
MAJOR_TWENTYFOURTH = .2
PERFECT_TWENTYFIFTH = .55
AUGMENTED_TWENTYFIFTH = .65
PERFECT_TWENTYSIXTH = .1
MINOR_TWENTYSEVENTH = .275
MAJOR_TWENTYSEVENTH = .25
MINOR_TWENTYEIGHTH = .7
MAJOR_TWENTYEIGHTH = 0.9
PERFECT_TWENTYNINTH = 0

EXITS_AND_ENTRANCES = []
NEW_ENTRANCE_TIME = 0
SIMULTANEOUS_EVENTS = []
EVENT = []
OCTAVE_SEPARATION = .02
DOWNBEAT = "downbeat"

METRIC_TENSION_TABLE = [[4, [1, 2], [2, 2], [3, 6], [4, 2]],
                        [2, [1, 2], [2, 2]],
                        [3, [1, 2], [2, 2], [3, 2]],
                        [6, [1, 2], [2, 2], [3, 2], [4, 8], [5, 4], [6, 3]],
                        [9, [1, 2], [2, 2], [3, 2], [4, 8], [5, 4], [6, 3], [7, 14], [8, 8], [9, 4]]]

INTERVAL_LIST = [UNISON, MINOR_SECOND, MAJOR_SECOND, MINOR_THIRD, MAJOR_THIRD,
                 PERFECT_FOURTH, AUGMENTED_FOURTH, PERFECT_FIFTH, MINOR_SIXTH,
                 MAJOR_SIXTH, MINOR_SEVENTH, MAJOR_SEVENTH, PERFECT_OCTAVE, MINOR_NINTH,
                 MAJOR_NINTH, MINOR_TENTH, MAJOR_TENTH, PERFECT_ELEVENTH,
                 AUGMENTED_ELEVENTH, PERFECT_TWELVETH, MINOR_THIRTEENTH, MAJOR_THIRTEENTH,
                 MINOR_FOURTEENTH, MAJOR_FOURTEENTH, PERFECT_FIFTEENTH, MINOR_SIXTEENTH,
                 MAJOR_SIXTEENTH, MINOR_SEVENTEENTH, MAJOR_SEVENTEENTH, PERFECT_EIGHTEENTH,
                 AUGMENTED_EIGHTEENTH, PERFECT_NINETEENTH, MINOR_TWENTIETH,
                 MAJOR_TWENTIETH, MINOR_TWENTYFIRST, MAJOR_TWENTYFIRST,
                 PERFECT_TWENTYSECOND, MINOR_TWENTYTHIRD, MAJOR_TWENTYTHIRD,
                 MINOR_TWENTYFOURTH, MAJOR_TWENTYFOURTH, PERFECT_TWENTYFIFTH,
                 AUGMENTED_TWENTYFIFTH, PERFECT_TWENTYSIXTH, MINOR_TWENTYSEVENTH,
                 MAJOR_TWENTYSEVENTH, MINOR_TWENTYEIGHTH, MAJOR_TWENTYEIGHTH,
                 PERFECT_TWENTYNINTH]

ROOT_STRENGHTS_AND_ROOTS = [[7, 0, 1], [5, 5, 2], [4, 0, 3], [8, 8, 4], [3, 0, 5], [9, 9, 6],
                            [2, 2, 7], [10, 0, 8], [1, 1, 9], [11, 0, 10], [0, 0, 11], [6, 6, 12]]


# (run-the-speac-weightings '((0 45 1000 4 55) (0 64 1000 3 55) (0 69 1000 2 55) . . . 4 8 4)
# (0.56 0.41 0.7799999999999999 0.51 1.33 0.51 1.26 0.51)
def run_the_speac_weightings(events, begin_beat, total_beats, meter):
    break_entrance = break_at_each_entrance(events)

    beat_lists_collect = collect_beat_lists(break_entrance)

    pitches = collect_pitch_lists(beat_lists_collect)

    vertical_tensions = create_lists_of_tensions(pitches)

    metric_tensions = map_metric_tensions(begin_beat, total_beats, meter)

    duration_tensions = compute_duration_tensions(events)

    approach_tensions = get_root_motion_weightings(events)

    return map_add(vertical_tensions, metric_tensions, duration_tensions, approach_tensions)


# (CREATE-LISTS-OF-TENSIONS (COLLECT-pitch-lists (collect-beat-lists (BREAK-AT-EACH-ENTRANCE book-example))))
# (0.3 0.3 0.5 0.3 0.5 0.3 0.3 0.3)
def create_lists_of_tensions(groups_of_midi_notes):
    groups_to_intervals = translate_groups_to_intervals(groups_of_midi_notes)
    rates = rate_lists(groups_to_intervals)
    result = compare(rates)
    return result


# (translate-groups-to-intervals '(((73 69 64 45)) ((73 69 64 57)  . . .
# (((19 28)) ((7 16) (7 17)) ((8 15)) . . .
def translate_groups_to_intervals(groups_of_midi_notes):
    result = []
    for beat in groups_of_midi_notes:
        result.append(translate_to_intervals(beat))
    return result


# (translate-to-intervals '((73 69 64 45)))
# ((19 28))
def translate_to_intervals(groups_of_midi_notes):
    result = []
    for list in groups_of_midi_notes:
        list_res = []
        list.sort()
        remove_octaves_res.clear()
        arranged_midi_notes = remove_octaves(list)
        bass_note = arranged_midi_notes[0]
        for note in arranged_midi_notes:
            if note != bass_note:
                list_res.append(note - bass_note)

        result.append(list_res)
    return result


remove_octaves_res = []


# (remove-octaves '(60 67 64 72 60 67 76))
# (60 67 64)
def remove_octaves(notes):
    if len(notes) == 0:
        return notes
    remove_octaves_res.append(notes[0])
    remove_octaves(remove_note(notes[0], notes[0:]))
    return remove_octaves_res


# (remove-note 60 '(67 64 72))
# (67 64)
def remove_note(note, notes):
    result = list(filter(lambda x: x % 12 != note % 12, notes))
    return result


# (collect-pitch-lists '((((0 73 1000 1 55) (0 69 1000 2 55) . . .
# (((73 69 64 45)) ((73 69 64 57) . . .
def collect_pitch_lists(pitch_lists):
    result = []
    for pitch_list in pitch_lists.copy():
        result.append(collect_pitches(pitch_list))
    return result


# (collect-pitches '(((6000 73 1000 1 55) (6000 64 1000 2 55) (6000 57 1000 3 55) (6000 57 1000 4 55))))
# ((73 64 57 57))
def collect_pitches(lists):
    result = []
    for events in lists:
        result.append(collect_pitch(events))
    return result


# (collect-pitch '((6000 73 1000 1 55) (6000 64 1000 2 55) (6000 57 1000 3 55) (6000 57 1000 4 55)))
# (73 64 57 57)
def collect_pitch(events):
    result = []
    for event in events:
        result.append(event[1])
    return result


# Сравнивает рейтинги, чтобы найти более созвучные для каждого бита
# (compare '((0.3) (0.3 0.65) (0.5) (0.3) (0.92 0.5) (0.3 1.2) (0.3) (0.3)))
# (0.3 0.3 0.5 0.3 0.5 0.3 0.3 0.3)
def compare(beat_rating_lists):
    result = []
    for ratings in beat_rating_lists:
        result.append(min(ratings))
    return result


# Назначает веса спискам
# (rate-lists '(((19 28)) ((7 16) (7 17)) ((8 15))  . . .
# ((0.3) (0.3 0.65) (0.5) (0.3) (0.92 0.5) (0.3 1.2) (0.3) (0.3))
def rate_lists(lists_of_intervals):
    result = []
    for list in lists_of_intervals:
        result.append(rate(list))
    return result


# Преобразует аргумент в вес на основе сохранённых значений
# (rate '((7 16) (7 16)))
# (0.3 0.3)
def rate(lists_of_intervals):
    lists_of_intervals = copy.deepcopy(lists_of_intervals)
    result = []

    while True:
        if len(lists_of_intervals) == 0:
            return result
        else:
            first_list = lists_of_intervals.pop(0)
            intervals_rate = rate_the_intervals(first_list)

            list_result = 0.0
            for rate in intervals_rate:
                list_result += rate

            rounded_result = my_round(list_result)
            result.append(rounded_result)


# (my-round 1.566)
#    1.57
def my_round(number):
    return float((round(number * 100)) / 100)


# Вохвращает рейтинг для входного интервала
# (rate-the-intervals '(7 16))
# ((0.1 0.2))
def rate_the_intervals(list_of_intervals):
    list_of_intervals = copy.deepcopy(list_of_intervals)
    result = []

    while True:
        if len(list_of_intervals) == 0:
            return result
        else:
            first_list = list_of_intervals.pop(0)
            result.append(INTERVAL_LIST[first_list])


# (collect-beat-lists '(((0 73 1000 1 55) (0 69 1000 2 55) . . .
# ((((0 73 1000 1 55) (0 69 1000 2 55) . . .
def collect_beat_lists(entrance_lists):
    result = []
    local_result = []
    need_to_add_next = False
    first = True
    for simultaneous_event in entrance_lists:
        if not need_to_add_next and not first:
            result.append(local_result)
            local_result = []

        first = False
        local_result.append(simultaneous_event)
        need_to_add_next = False

        for event in simultaneous_event:
            if len(event) == 6:
                need_to_add_next = True

    result.append(local_result)
    return result


def group_bits(entrance_lists):
    return [entrance_lists]


# (break-at-each-entrance ((0 45 1000 4 55) (0 64 1000 3 55)  . . .
# (((0 73 1000 1 55) (0 69 1000 2 55)  . . .
def break_at_each_entrance(events):
    exits_and_entrances = []
    events = fix_the_triplets(events)
    events.sort()

    while len(events) != 0:
        event = events[0]
        simultaneous_events = place_events_in_channel_order(collect_simultaneous_events(events))

        if exits_and_entrances != []:
            new_entrance_time = get_new_exit_and_entrance_time(events, event[0])
        else:
            new_entrance_time = get_new_entrance_time(events, event[0], event[0] + event[2])

        reset_dur = reset_durations(new_entrance_time, simultaneous_events)
        exits_and_entrances.append(reset_dur)

        reset_next_dur = reset_next_durations(new_entrance_time, copy.deepcopy(simultaneous_events))
        remove_zero_dur = remove_zero_durations(reset_next_dur)
        if remove_zero_dur != []:
            for element in remove_zero_dur:
                events.append(element)
        remove_a = remove_all(simultaneous_events, events)
        events.sort()

    return exits_and_entrances


# (place-events-in-channel-order '((0 73 1000 2 55) (0 69 1000 1 55)))
# ((0 69 1000 1 55) (0 73 1000 2 55))
def place_events_in_channel_order(events):
    ordered_channels = []
    ordered_events = []

    for event in events:
        ordered_channels.append(event[3])
    ordered_channels.sort()

    events_copy = copy.deepcopy(events)

    while len(ordered_events) != len(ordered_channels):
        for channel in ordered_channels:
            for event in events_copy:
                if event[3] == channel:
                    ordered_events.append(event)
                    events_copy.remove(event)
                    break
    return ordered_events


# Выводит нужный элемент бита
# (find-channel-event 2 '((0 73 1000 1 55) (0 69 1000 2 55) (0 64 1000 3 55) (0 45 1000 4 55)))
# ((0 69 1000 2 55)
def find_channel_event(channel, events):
    return events[channel - 1]


# (get-channel 1 '((0 45 1000 1 55) (0 64 1000 3 55)))
# ((0 45 1000 1 55))
def get_channel(number, music):
    channel_music = []
    for event in music:
        if event[3] == number:
            channel_music.append(event)
    return channel_music


# (get-all-channels '((0 45 1000 4 55) (0 64 1000 3 55) . . .
# (((0 73 1000 1 55) (1000 73 500 1 55)  . . .
def get_all_channels(music):
    all_channels_music = []
    for i in range(1, 17):
        one_channel_music = get_channel(i, music)
        all_channels_music.append(one_channel_music)
    return all_channels_music


# (remove-zero-durations '((1000 72 0 1 127) (1000 67 0 2 127) (1000 64 0 3 127) (1000 60 0 4 127)))
#  nil
def remove_zero_durations(events):
    return list(filter(lambda event: event[2] != 0, events))


# Собирает в один список ноты, которые играются одновременно в первый момент времени
# (collect-simultaneous-events '((0 73 1000 1 55) (0 69 1000 2 55)  . . .
# ((0 73 1000 1 55) (0 69 1000 2 55) (0 64 1000 3 55) (0 45 1000 4 55))
def collect_simultaneous_events(events):
    first_bit_notes = []
    one_time = events[0][0]
    for event in events:
        if event[0] != one_time:
            break
        else:
            first_bit_notes.append(event)
    return first_bit_notes


# (fix-triplets '((0 73 1000 1 55) (1001 73 500 1 55) (1000 74 500 1 55) (999 74 500 55)))
# ((0 73 1001 1 55) (1001 73 500 1 55) (1000 74 500 1 55) (999 74 500 55))
def fix_triplets(events):
    new_events = []
    for i in range(0, len(events)):
        cur_event = events[i]
        # если есть следующая нота
        if i != len(events) - 1:
            next_event = events[i + 1]
            # если следующая нота начинается на 1 раньше или позже
            # (то есть, например, вместо 1000 999 или 1001), то длительность предыдущей ноты
            # изменяется на 1
            if abs(cur_event[0] + cur_event[2] - next_event[0]) == 1:
                cur_event[2] -= (cur_event[0] + cur_event[2] - next_event[0])

        new_events.append(cur_event)
    return new_events


# (fix-the-triplets '((0 45 1000 4 55) (0 64 1000 3 55) (0 69 1000 2 55) (0 73 1000 1 55) (1000 57 1000 4 55) . . .
# ((0 73 1000 1 55) (1000 73 500 1 55)  . . .
def fix_the_triplets(events):
    res = []
    for channel in get_all_channels(events):
        if channel != []:
            triplets_fix = fix_triplets(channel)
            for event in triplets_fix:
                res.append(event)
    return res


# (remove-all '((0 73 333 1 55)) '((0 73 333 1 55) (0 69 444 2 55) (1500 45 1000 4 55)))
# ((0 69 444 2 55) (1500 45 1000 4 55))
def remove_all(simultaneous_events, ordered_events):
    for event in simultaneous_events:
        ordered_events.remove(event)
    return ordered_events


# Вернёт или мин. длительность звучания ноты в ивенте или время начала следующего ивента
# (get-new-exit-and-entrance-time '((0 73 333 1 55) (0 69 444 2 55) (1500 45 1000 4 55)) 0)
# 333
def get_new_exit_and_entrance_time(events, start_time):
    events.sort()
    end_time = get_shortest_duration(start_time, events) + start_time
    new_start_time = get_next_start_time(start_time, events)

    if end_time is None:
        return new_start_time
    elif new_start_time is None:
        return end_time
    elif end_time > new_start_time:
        return new_start_time
    else:
        return end_time


# (get-next-start-time 0 '((0 73 1000 1 55) (0 69 1000 2 55)  . . .
# 1000
def get_next_start_time(start_time, events):
    for event in events:
        if event[0] != start_time:
            return event[0]
    return None


# (get-new-entrance-time '((0 73 1000 1 55) (0 69 1000 2 55) (0 64 1000 3 55) (0 45 1000 4 55) (1000 73 500 1 55) . . .
# 1000
def get_new_entrance_time(events, start_time, end_time):
    for event in events:
        if event[0] != start_time:
            return event[0]
    return end_time


# (get-shortest-duration 0 '((0 73 1000 1 55) (0 69 1000 2 55)  . . .
# 1000
def get_shortest_duration(start_time, events):
    start_time_events = []
    for event in events:
        if event[0] != start_time:
            break
        else:
            start_time_events.append(event)
    return get_shortest(start_time_events)


# (get-shortest '((0 73 1000 1 55) (0 69 1000 2 55) (0 64 1000 3 55) (0 45 1000 4 55)))
#    1000
def get_shortest(events):
    durations = []
    for event in events:
        durations.append(event[2])
    durations.sort()
    return durations[0]


# (reset-durations 1000 '((0 73 1500 1 55) (0 69 1000 2 55) (0 64 1000 3 55) (0 45 1000 4 55)))
# ((0 73 1000 1 55 *) (0 69 1000 2 55) (0 64 1000 3 55) (0 45 1000 4 55))
def reset_durations(new_entrance_time, simultaneous_events):
    edited_events = []
    for event in simultaneous_events:
        new_event = [event[0], event[1]]
        need_to_add_char = False
        if event[0] + event[2] > new_entrance_time:
            new_event.append(new_entrance_time - event[0])
            need_to_add_char = True
        else:
            new_event.append(event[2])
        new_event.append(event[3])
        new_event.append(event[4])
        if need_to_add_char:
            new_event.append("*")
        edited_events.append(new_event)
    return edited_events


# (reset-next-durations 1000 '((0 73 1000 1 55) (0 69 1000 2 55) (0 64 1000 3 55) (0 45 1000 4 55)))
# ((1000 73 0 1 55) (1000 69 0 2 55) (1000 64 0 3 55) (1000 45 0 4 55))
def reset_next_durations(new_entrance_time, ordered_events):
    edited_events = []
    for event in ordered_events:
        event[2] = event[2] - (new_entrance_time - event[0])
        event[0] = new_entrance_time
        edited_events.append(event)
    return edited_events


# (map-metric-tensions 4 8 4)
# (0.2 0.05 0.1 0.05 0.2 0.05 0.1 0.05)
def map_metric_tensions(start_beat, total_beats, meter, beat=0, result=[]):
    start_beat_copy = copy.deepcopy(start_beat)
    total_beats = copy.deepcopy(total_beats)

    if beat == 0:
        result.clear()

    if total_beats != beat:

        result.append(lookup_and_figure_metric_tension(meter, start_beat_copy))

        if start_beat_copy == meter:
            new_start_beat = 1
        else:
            new_start_beat = start_beat_copy + 1

        map_metric_tensions(new_start_beat, total_beats, meter, beat + + 1)
    return result


# (lookup-and-figure-metric-tension 9 7)
# 0.05
def lookup_and_figure_metric_tension(meter, beat_number):
    for element in METRIC_TENSION_TABLE:
        if element[0] == meter:
            return float('{:.3}'.format((beat_number * 0.1) / (element[beat_number][1])))


# Добавление карт по различным параметрам анализа
# (map-add '(0.3 0.3 0.5 0.3 0.5 0.3 0.3 0.3)
# '(0.2 0.05 0.1 0.05000000000000001 0.2 0.05 0.1 0.05000000000000001)
# '(0.06 0.06 0.08 0.06 0.08 0.06 0.06 0.06) (0 0 0.1 0.1 0.55 0.1 0.8 0.1))
# (0.56 0.41 0.7799999999999999 0.51 1.33 0.51 1.26 0.51)
def map_add(list_1, list_2, list_3, list_4):
    parameters_result = []
    for i in range(1, len(list_1) + 1):
        if len(list_1) < i or len(list_2) < i or len(list_3) < i or len(list_4) < i:
            break
        cur_res = numpy.float32(list_1[i - 1])
        cur_res += numpy.float32(list_2[i - 1])
        cur_res += numpy.float32(list_3[i - 1])
        cur_res += numpy.float32(list_4[i - 1])
        parameters_result.append(cur_res)
    return parameters_result


# Returns the interval between ontimes
# (get-durations '(0 1000 2000 3000 4000 5000 6000 7000))
#  (1000 1000 1000 1000 1000 1000 1000 1000)
def get_durations(ontimes):
    result = []

    before = None
    for i in range(1, len(ontimes) + 1):
        if before is None:
            before = ontimes[i - 1]
        else:
            result.append(ontimes[i - 1] - before)
            before = ontimes[i - 1]

    if len(ontimes) == 1 or len(ontimes) == 0:
        result.append(0)
    else:
        result.append(result[len(result) - 1])

    return result


# (duration-map '((((0 73 1000 1 55) (0 69 1000 2 55) (0 64 1000 3 55) (0 45 1000 4 55))) (((1000 73 500 1 55) . . .
# (1000 1000 1000 1000 1000 1000 1000 1000)
def duration_map(beats):
    ontimes = []
    for beat in beats:
        ontimes.append(beat[0][0][0])
    return get_durations(ontimes)


# (compute-duration-tensions book-example)
# (0.06 0.06 0.08 0.06 0.08 0.06 0.06 0.06)
def compute_duration_tensions(events):
    each_entrance_break = break_at_each_entrance(events)

    beat_lists_collect = collect_beat_lists(each_entrance_break)

    durations = duration_map(beat_lists_collect)

    pitch_lists_collect = collect_pitch_lists(beat_lists_collect)
    interval_tensions = create_lists_of_tensions(pitch_lists_collect)

    result = []
    i = 0
    # В LISP числа с плавающей точкой 32-разрядные, в Python - нет.
    # Чтобы результаты совпадали приводим к этому формату
    for duration in durations:
        local_res = numpy.float32(0.01) * numpy.float32(
            round(numpy.float32(100) * ((numpy.float32((duration / 4000)) * numpy.float32(0.1)) +
                                        (numpy.float32(interval_tensions[i]) * numpy.float32(0.1)))))
        result.append(local_res)

        i += 1
    return result


# (get-root-motion-weightings book-example)
# (0 0 0.1 0.1 0 0.8 0.8 0.1)
def get_root_motion_weightings(events):
    roots = get_chord_roots(events)
    result = find_motion_weightings(roots)
    result.insert(0, 0)
    return result


# (find-motion-weightings '(45 57 64 57 62 55 57 50))
#  (0 0.1 0.1 0.55 0.1 0.8 0.1)
def find_motion_weightings(roots):
    size = len(roots)
    result = []
    roots_copy = roots.copy()
    for i in range(1, size):
        cur_root = roots_copy[0]

        if len(roots_copy) > 1:
            next_root = roots_copy[1]
            diff = abs(cur_root - next_root)
            result.append(INTERVAL_LIST[diff % 12])
            roots_copy.pop(0)

    return result


# (get-chord-roots book-example)
# (45 57 64 57 62 55 57 50)
def get_chord_roots(events):
    pitch_list = collect_pitch_lists(collect_beat_lists(break_at_each_entrance(events)))
    on_beat_pitch_lists = []

    for i in range(1, len(pitch_list) + 1):
        scrunch_local_res = scrunch(pitch_list[i - 1])
        scrunch_local_res.sort()
        on_beat_pitch_lists.append(scrunch_local_res)

    intervals = []
    for element in on_beat_pitch_lists:
        element_copy = element.copy()
        intervals.append(derive(element_copy))

    roots = []
    for element in intervals:
        element_copy = element.copy()
        roots.append(find_strongest_root_interval(element_copy))

    result = []
    for root in roots:
        interval_in_chord = find_interval_in_chord(root, on_beat_pitch_lists[0])
        interval_in_chord.sort()
        upper_lower_res = find_upper_lower(root, interval_in_chord)
        result.append(upper_lower_res)
        on_beat_pitch_lists.pop(0)

    return result


# (derive (55 64 84)) -> (0 5 8 9)
# (derive (72))       -> (0)
# (derive (41 74 76)) -> (0 2 9 11)
def derive(pitches):
    intervals = derive_all_intervals(pitches)
    return sorted(set(intervals))


# (derive-all-intervals '(45 64 69 73))
#  (0 7 0 4 0 5 9 0 4)
def derive_all_intervals(pitches):
    result = []
    size = len(pitches)

    for i in range(1, size):
        local_result = derive_intervals(pitches)
        for e in local_result:
            result.append(e)
        pitches.pop(0)
    result.append(0)

    return result


# (derive-intervals '(45 64 69 73))
#   (0 7 0 4)
def derive_intervals(pitches):
    result = []
    first_pitch = pitches[0]

    for pitch in pitches:
        diff = pitch - first_pitch
        result.append(diff % 12)
    return result


def scrunch(lists):
    result = []
    for list in lists:
        for element in list:
            if element not in result:
                result.append(element)
    return result


# Returns the root note
# (find-upper-lower 7 '(45 64))
#  45
def find_upper_lower(root, interval):
    test = None
    for element in ROOT_STRENGHTS_AND_ROOTS:
        if root == element[0]:
            test = element
            break
    if test[1] == 0:
        return interval[0]
    else:
        return interval[1]


# (find-strongest-root-interval '(0 7 4 0 0 9 5 0 8))
#   7
def find_strongest_root_interval(intervals):
    intervals_copy = copy.deepcopy(intervals)
    test = find_root_strenghts_and_roots(intervals_copy)
    sorted_test = sort_CADDR(test)
    return sorted_test[0][0]


# ('((2 (a b c))(3 (d e f))(1 (g h i))))
#  >> ((1 (G H I)) (2 (A B C)) (3 (D E F)))
def sort_CADDR(lists):
    lists.sort(key=sort_by_third_element)
    return lists


# Внешняя функция, которая будет сортировать списки по 3 элементу
def sort_by_third_element(some_list):
    return some_list[2]  # Ключом является третий символ в каждом списке, сортируем по нему


# (find-root-strengths-and-roots '(0 7 4 0 0 9 5 0 8))
#   ((0 0 11) (7 0 1) (4 0 3) (0 0 11) (0 0 11) (9 9 6) (5 5 2) (0 0 11) (8 8 4))
def find_root_strenghts_and_roots(pc_intervals):
    result = []
    for interval in pc_intervals:
        for element in ROOT_STRENGHTS_AND_ROOTS:
            if interval == element[0]:
                result.append(element)
                break
    return result


# (find-interval-in-chord 7 '(54 62 64 69))
# (69 62)
def find_interval_in_chord(interval, chord):
    if len(chord) == 1:
        return chord
    else:
        return find_it_in_chord(interval, derive_all_pitches(chord))


# (find-it-in-chord 7 '(54 54 54 62 54 64 54 69 62 62 62 64 62 69 64 64 64 69))
#  (69 62)
def find_it_in_chord(interval, chord):
    size = len(chord)
    for i in range(1, size):
        pitch = chord[0]
        if len(chord) > 1:
            next_pitch = chord[1]
            diff = next_pitch - pitch
            if diff % 12 == interval:
                return [next_pitch, pitch]
        chord.pop(0)
    # В коде на LISP происходит чудо, там (mod -5 11) даст результат 6 (типа 11 - 5)...
    return None


# (derive-all-pitches '(54 62 64 69))
# (54 54 54 62 54 64 54 69 62 62 62 64 62 69 64 64 64 69)
def derive_all_pitches(pitches):
    result = []
    size = len(pitches)
    for i in range(1, size):
        pitch = pitches[0]
        for two_pitches in derive_pitches(pitches):
            for e in two_pitches:
                result.append(e)
        pitches.remove(pitch)

    return result


# (derive-pitches '(54 62 64 69))
#  ((54 54) (54 62) (54 64) (54 69))
def derive_pitches(pitches):
    result = []
    first_pitch = pitches[0]
    for pitch in pitches:
        local_result = [first_pitch, pitch]
        result.append(local_result)
    return result



import copy

from .speac import get_channel


def pattern_match(pattern_1, pattern_2, number_wrong_possible, speac_settings):
    pattern_1 = copy.deepcopy(pattern_1)
    pattern_2 = copy.deepcopy(pattern_2)

    if pattern_1 == [] and pattern_2 == []:
        return True

    elif number_wrong_possible == -1 \
            or abs(pattern_1[0] - pattern_2[0]) > speac_settings.AMOUNT_OFF:
        return False

    elif pattern_1[0] != pattern_2[0]:
        pattern_1.pop(0)
        pattern_2.pop(0)
        return pattern_match(pattern_1, pattern_2, number_wrong_possible - 1, speac_settings)

    else:
        pattern_1.pop(0)
        pattern_2.pop(0)
        return pattern_match(pattern_1, pattern_2, number_wrong_possible, speac_settings)


def get_ontimes_and_pitches(events):
    result = []

    while True:
        if len(events) == 0:
            break
        else:
            event = events.pop(0)
            result.append([event[0], event[1]])
    return result


def run_pattern_match(pattern, patterns, speac_settings):
    matches = 0

    while True:

        first_n_patterns = patterns[:len(pattern)]

        if len(patterns) < len(pattern):
            return matches

        else:
            if pattern_match(pattern, first_n_patterns, speac_settings.INTERVALS_OFF, speac_settings):
                patterns.pop(0)
                matches += 1

            else:
                patterns.pop(0)


def interval_translator(midi_list):
    result = []

    while True:
        if len(midi_list) == 1:
            break

        else:
            result.append(midi_list[1] - midi_list[0])
            midi_list.pop(0)

    return result


DURATION = "no"


def find_matchings(pattern, patterns, speac_settings):
    translator_input = []
    for p in pattern:
        if DURATION == "yes":
            translator_input.append(p[2])
        else:
            translator_input.append(p[1])

    translator_res = interval_translator(translator_input)

    translator_input = []
    for p in patterns:
        if DURATION == "yes":
            translator_input.append(p[2])
        else:
            translator_input.append(p[1])
    translator_res_2 = interval_translator(translator_input)

    result = run_pattern_match(translator_res, translator_res_2, speac_settings)

    return result


def firstn(number, some_list):
    if len(some_list) < number:
        return firstn(number - 1, some_list)
    else:
        return butlast(some_list, len(some_list) - number)


def butlast(some_list, number=1):
    result = copy.deepcopy(some_list)

    try:
        for i in range(1, number + 1):
            result.pop(len(result) - 1)
        return result
    except IndexError:
        return []


def find_the_matches(work_1, work_2, speac_settings):
    result = []

    while True:
        if len(work_1) < speac_settings.PATTERN_SIZE:
            break

        else:
            patterns = firstn(speac_settings.PATTERN_SIZE, work_1)
            other_patterns = work_1[speac_settings.PATTERN_SIZE:]
            test = find_matchings(patterns, work_2, speac_settings)
            # print(test)

            if test > speac_settings.THRESHOLD:
                translator_input = []
                for p in patterns:
                    translator_input.append(p[1])
                local_result = [test, patterns[0][0], interval_translator(translator_input)]
                result.append(local_result)
                work_1 = other_patterns

            else:
                work_1.pop(0)

    return result


def sort_by_first_element(some_list):
    return some_list[0]


def simple_matcher(events, speac_settings):
    # print("events = ", events)
    events.sort(key=sort_by_first_element)
    # print("events = ", events)
    channel = get_channel(speac_settings.MATCHING_LINE, events)
    # print("channel = ", channel)
    ordered_and_channeled_events = get_ontimes_and_pitches(channel)
    # print("ordered = ", ordered_and_channeled_events)
    result = find_the_matches(ordered_and_channeled_events, ordered_and_channeled_events, speac_settings)
    return result


import copy

from .pattern_match import pattern_match, simple_matcher
from .speac import remove_all
from .speac_analysis import capture_beats

c1 = [24, 36, 48, 60, 72, 84, 96, 108, 27, 28, 39, 40, 51, 52, 63, 64, 75, 76, 87, 88, 99, 100,
      31, 43, 55, 67, 79, 91, 103]
p1 = [29, 41, 53, 65, 77, 89, 101, 33, 93, 105, 81, 45, 57, 69, 24, 36, 48, 60, 72, 84, 96, 108]
a1 = [31, 43, 55, 67, 79, 91, 103, 35, 47, 59, 71, 83, 95, 107, 26, 38, 50, 62, 74, 86, 98, 29,
      41, 53, 65, 77, 89, 101]
a2 = [35, 47, 59, 71, 83, 95, 107, 26, 38, 50, 62, 74, 86, 98, 29, 41, 53, 65, 77, 89, 101]
c4 = [28, 40, 52, 64, 76, 88, 100, 31, 43, 55, 67, 79, 91, 103, 35, 47, 59, 71, 83, 95, 107]
p2 = [26, 38, 50, 62, 74, 86, 98, 89, 101, 77, 65, 29, 41, 53, 93, 105, 81, 69, 33, 45, 57]
c2 = [33, 45, 57, 69, 81, 93, 105, 24, 36, 48, 60, 72, 84, 96, 108, 28, 40, 52, 64, 76, 88, 100]
s1 = [26, 38, 50, 62, 74, 86, 98, 30, 33, 42, 54, 66, 78, 90, 102, 45, 57, 69, 81, 93, 105]
s3 = [28, 40, 52, 64, 76, 88, 100, 32, 35, 44, 56, 68, 80, 92, 104, 47, 59, 71, 83, 95, 107]
e1 = [33, 45, 57, 69, 81, 93, 105, 25, 37, 49, 61, 73, 85, 97, 28, 40, 52, 64, 76, 88, 100]
e3 = [35, 47, 59, 71, 83, 95, 107, 27, 39, 51, 63, 75, 87, 99, 30, 42, 54, 66, 78, 90, 102]
c3 = [24, 36, 48, 60, 72, 84, 96, 108, 28, 31, 34, 40, 52, 64, 76, 88, 100, 43, 55, 67, 79, 91, 103,
      46, 58, 70, 82, 94, 106]
e2 = [25, 37, 49, 61, 73, 85, 97, 28, 31, 34, 40, 52, 64, 76, 88, 100, 43, 55, 67, 79, 91, 103, 46,
      58, 70, 82, 94, 106]
e4 = [27, 39, 51, 63, 75, 87, 99, 24, 30, 33, 36, 48, 60, 72, 84, 96, 108, 42, 54, 66, 78, 90, 102,
      45, 57, 69, 81, 93, 105]
a3 = [32, 44, 56, 68, 80, 92, 104, 26, 29, 35, 38, 50, 62, 74, 86, 98, 41, 53, 65, 77, 89, 101, 47,
      59, 71, 83, 95, 107]
p3 = [32, 44, 56, 68, 80, 92, 104, 24, 30, 36, 48, 60, 72, 84, 96, 108, 42, 54, 66, 78, 90, 102]
p4 = [25, 37, 49, 61, 73, 85, 97, 29, 32, 41, 53, 65, 77, 89, 101, 44, 56, 68, 80, 92, 104]
s4 = [30, 42, 54, 66, 78, 90, 102, 25, 37, 49, 61, 73, 85, 97, 34, 46, 58, 70, 82, 94, 106]
a4 = [27, 39, 51, 63, 75, 87, 99, 31, 43, 55, 67, 79, 91, 103, 34, 46, 58, 70, 82, 94, 106]
s2 = [34, 46, 58, 70, 82, 94, 106, 26, 38, 50, 62, 74, 86, 98, 29, 41, 53, 65, 77, 89, 101]

ANALYSIS_LEXICON = [c1, p1, a1, a2, c4, p2, c2, s1, s3, e1, e3, c3, e2, e4, a3, p3, p4, s4, a4, s2]
ANALYSIS_LEXICON_NAMES = ["c1", "p1", "a1", "a2", "c4", "p2", "c2", "s1", "s3", "e1",
                          "e3", "c3", "e2", "e4", "a3", "p3", "p4", "s4", "a4", "s2"]


# Возвращает номер бита и соответсвующий ему SPEAC-символ
# (get-function '(0 (79 64 48 48 65 65 50 50)))
# (0 a1)
def get_function(beat_number, chord_notes):
    if len(chord_notes) < 2:
        return [beat_number, "e4"]
    else:
        return [beat_number, compare_them(chord_notes, ANALYSIS_LEXICON)[1]]


# По списоку нот возвращает значение подходящего SPEAC-символа
# (compare-them '(79 64 48 48 65 65 50 50) *analysis-lexicon*)
# ((31 43 55 67 79 91 103 35 47 59 71 83 95 107 26 38 50 62 74 86 98 29 41 53 65
#   77 89 101) a1)
def compare_them(harmonic_notes, harmonic_functions):
    highest = -1
    highest_index = -1
    counts = count_harmonic_notes(harmonic_notes, harmonic_functions)
    for i in range(0, len(counts)):
        if counts[i] > highest:
            highest = counts[i]
            highest_index = i

    return ANALYSIS_LEXICON[highest_index], ANALYSIS_LEXICON_NAMES[highest_index]


# Считает сколько элементов из первого списка содержится в каждом подсписке второго списка
# (count-harmonic-notes '(79 64 48 48 65 65 50 50) *analysis-lexicon*)
# (4 4 5 4 2 4 3 2 1 1 0 4 2 2 4 2 2 0 1 4)
def count_harmonic_notes(harmonic_notes, harmonic_functions):
    count = []
    for speac_symbol_list in harmonic_functions:
        count.append(my_count(harmonic_notes, speac_symbol_list))
    return count


# Считает сколько элементов из первого списка содержится во втором
# (my-count '(79 64 48 48 65 65 50 50) '(38 50 62 74 86 41 53 65 77 89 46 58 70 82 94))
# 4
def my_count(list1, list2):
    count = 0
    for element1 in list1:
        for element2 in list2:
            if element1 == element2:
                count += 1
    return count


def cadences(events, speac_settings):
    events = copy.deepcopy(events)
    captured_beats = capture_beats(events, speac_settings.BEAT)
    # print("captured_beats = ", captured_beats)

    pitches = []
    for beat in captured_beats:
        pitches.append(get_pitches(beat))

    functions = []
    for pitch in pitches:
        fun = get_function(pitch[0], pitch[1])
        functions.append(fun)

    best_cadences = return_best_cadences(functions, speac_settings)
    # print("best_cadences = ", best_cadences)

    result = []
    for element in best_cadences:
        reversed_element = [element[1], element[0]]
        result.append(reversed_element)

    return result


def next_4(lists):
    if (len(lists) >= 1 and lists[0][1].lower() == "c1") or \
            (len(lists) >= 2 and lists[1][1].lower() == "c1") or \
            (len(lists) >= 3 and lists[2][1].lower() == "c1") or \
            (len(lists) >= 4 and lists[3][1].lower() == "c1"):
        return True
    else:
        return False


def return_best_cadences(function_timing_lists, speac_settings, distance=0, previous=None, minor_flag=0):
    size = len(function_timing_lists)
    function_timing_lists = copy.deepcopy(function_timing_lists)
    result = []

    for i in range(1, size):
        function = function_timing_lists[0]
        next_function = function_timing_lists[1]
        function_timing_lists.pop(0)

        fun_1 = function[1].lower()

        if (minor_flag > 0) and (fun_1 == "c2") and (distance > speac_settings.CADENCE_MINIMUM):
            result.append([function[0], "c1"])
            distance = 0

        elif (minor_flag > 0) and (fun_1 == "c4") and (distance > speac_settings.CADENCE_MINIMUM):
            result.append([function[0], "a1"])
            distance = 0

        elif (previous == "a1") and (fun_1 == "c1") and (distance > speac_settings.CADENCE_MINIMUM):
            result.append(function)
            distance = 0

        elif (distance > speac_settings.CADENCE_MINIMUM) and (fun_1 == "a1") and (next_function[1].lower() != "c1"):
            result.append(function)
            distance = 0

        elif (fun_1 == "c1") and (distance > speac_settings.CADENCE_MINIMUM):
            result.append(function)
            distance = 0

        elif (fun_1 == "a1") and (not next_4(function_timing_lists)) and (distance > speac_settings.CADENCE_MINIMUM):
            result.append(function)
            distance = 0

        else:
            distance += 1000

        previous = fun_1
        minor_flag = set_minor_flag(function, minor_flag)

    return result


def set_minor_flag(list, flag):
    if list[1] == "s3" or list[1] == "a3" or list[1] == "p4":
        return 4
    else:
        if flag > 0:
            return flag - 1
        else:
            return 0


# (get-pitches '((61000 60 1000 4 96) (61000 76 1000 1 96) (61500 69 250 3 96)
#       (61500 72 500 2 96) (61750 70 250 3 96)))
# (61000 (60 76 69 72 70))
def get_pitches(events):
    result = [events[0][0]]
    pitch_result = []

    for event in events:
        pitch_result.append(event[1])

    result.append(pitch_result)
    return result


# (make-composite-rhythm *events*)
# (0 1000 1500 2000 2500 3000)
def make_composite_rhythm(events):
    events_set = set()
    for event in events:
        events_set.add(event[0])
    result = list(events_set)
    result.sort()
    return result


def collect_differential(ontime_map):
    previous = None
    count = 0
    time = ontime_map[0]
    result = []

    while len(ontime_map) >= 0:
        if len(ontime_map) == 0:
            result.append([count, time])

            if len(ontime_map) == 0:
                break

        elif previous is None:
            previous = ontime_map.pop(0)
            count += 1
            time = 0

        elif abs(ontime_map[0] - previous) > 1001:
            result.append([count, time])
            time = ontime_map[0]
            previous = ontime_map[0]
            count = 0

        else:
            previous = ontime_map.pop(0)
            count += 1

    return result


def composite_rhythm(events):
    comp_rhythm = make_composite_rhythm(events)
    return collect_differential(comp_rhythm)


def map_density(beats):
    result = []
    for beat in beats:
        result.append([len(beat), beat[0][0]])
    return result


def collect_by_differences(density_map):
    previous = None
    count = 0
    time = density_map[0][1]
    result = []

    while len(density_map) >= 0:
        if len(density_map) == 0:
            result.append([count, time])

            if len(density_map) == 0:
                break

        elif previous is None:
            previous = density_map.pop(0)
            count += 1

        elif abs(density_map[0][0] - previous[0]) > 2:
            result.append([count, time])
            time = density_map[0][1]
            previous = density_map[0]
            count = 0

        else:
            previous = density_map.pop(0)
            count += 1

    return result


def density(events, speac_settings):
    beats = capture_beats(events, speac_settings.BEAT)
    return collect_by_differences(map_density(beats))


def collect_patterns(list, lists, type, speac_settings):
    result = []

    while True:
        if len(lists) == 0:
            break

        elif pattern_match(list[2], lists[0][2], speac_settings.INTERVALS_OFF, speac_settings):
            local_result = []
            lists[0].pop(0)
            local_result.append(type)
            local_result.append(lists[0][0])
            local_result.append(lists[0][1])
            result.append(local_result)
            lists.pop(0)

        else:
            result.append(lists[0])
            lists.pop(0)

    return result


def almost_the_same_lists(lists, speac_settings):
    types = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k"]
    result = []

    while True:
        if len(lists) == 0:
            break

        elif not isinstance(lists[0][0], int):
            result.append(lists.pop(0))

        else:
            list_0_before = copy.deepcopy(lists[0])
            lists[0].pop(0)

            if len(types) != 0:
                types_0 = types[0]
            else:
                types_0 = []

            result.append([types_0, lists[0][0], lists[0][1]])

            if len(types) != 0:
                first_type = types.pop(0)
            else:
                first_type = []

            lists.pop(0)
            lists = collect_patterns(list_0_before, lists, first_type, speac_settings)
    return result


def evaluate_forms(max, min, forms):
    result = []

    while True:
        if len(forms) == 0:
            break

        elif (len(forms[0]) <= max) and (len(forms[0]) >= min):
            first_form = forms.pop(0)
            result.append(first_form)

        else:
            forms.pop(0)

    return result


def find_letters_used(form):
    while True:
        if len(form) == 0:
            return LETTERS_USED

        elif form[0][0] in ["a", "b", "c", "d", "e", "f", "g"]:
            if form[0][0] not in LETTERS_USED:
                LETTERS_USED.append(form[0][0])

        form.pop(0)


LETTERS = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m",
           "n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z"]


def name_them(forms, letters):
    result = []

    while True:
        if len(forms) == 0:
            return result
        else:
            if forms[0][0] in LETTERS:
                result.append(forms.pop(0))
            else:
                first_form = forms.pop(0)
                result.append([letters[0], first_form[1]])


def within_range(min, max, lists):
    lists = copy.deepcopy(lists)
    result = []

    while True:
        if len(lists) == 0:
            return result
        else:
            if (lists[0][1] >= min) and (lists[0][1] <= max):
                lists.pop(0)
            else:
                result.append(lists.pop(0))


def return_within_range(original, other, meter, speac_settings):
    amount = meter * speac_settings.BEAT * speac_settings.MEASURES

    while True:
        if len(original) == 0:
            return other
        else:
            first_original = original.pop(0)
            other = within_range(first_original[1] - amount, first_original[1] + amount, other)


def reduce_out_close_calls(pattern_discovery, forms, meter, speac_settings):
    result = []
    forms = copy.deepcopy(forms)
    pattern_discovery = copy.deepcopy(pattern_discovery)

    while True:
        if len(forms) == 0:
            return result
        else:
            first_form = forms.pop(0)
            result.append(return_within_range(pattern_discovery, first_form, meter, speac_settings))


def my_last(some_list):
    try:
        last = [some_list[-1]]
        first_of_last = last[0]
        return first_of_last
    except IndexError:
        return []


def butlast(some_list, number=1):
    result = copy.deepcopy(some_list)

    try:
        for i in range(1, number + 1):
            result.pop(len(result) - 1)
        return result
    except IndexError:
        return []


def combine(forms, meter, speac_settings):
    forms = copy.deepcopy(forms)

    reduce_result = reduce_out_close_calls(my_last(forms), butlast(forms), meter, speac_settings)

    input = []
    for element in forms[-1]:
        input.append(element)

    for element in reduce_result:
        for e in element:
            if e:
                input.append(e)

    used_letters = find_letters_used(my_last(forms))
    used_letters.reverse()

    alphabet = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m",
                "n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z"]

    letters = remove_all(used_letters, alphabet)

    result = name_them(input, letters)
    result.sort(key=sort_by_second_element)
    return result


def sort_by_second_element(some_list):
    return some_list[1]


LETTERS_USED = []


def eval_combine_and_integrate_forms(events, meter, speac_settings):
    global LETTERS_USED
    LETTERS_USED = []
    events = copy.deepcopy(events)
    cadences_result = cadences(events, speac_settings)
    # print("cadences_result = ", cadences_result)

    density_result = density(events, speac_settings)
    # print("density_result = ", density_result)

    rhythm_result = composite_rhythm(events)
    # print("rhythm_result = ", rhythm_result)

    s_matcher = simple_matcher(events, speac_settings)
    # print("s_matcher = ", s_matcher)

    patterns = almost_the_same_lists(s_matcher, speac_settings)
    # print("patterns = ", patterns)

    min = round((events[-1][0] + events[-1][2]) / (meter * 1000 * 8))
    max = round((events[-1][0] + events[-1][2]) / (meter * 1000 * 4))

    forms_result = evaluate_forms(max, min, [cadences_result, density_result, rhythm_result])
    # print("forms_res = ", forms_result)
    forms_result.append(patterns)
    return combine(forms_result, meter, speac_settings)


from setuptools import setup, find_packages

setup(
    name='speac',
    version='1.0.10',
    packages=find_packages(),
    url='https://github.com/GolzitskyNikolay/SPEAC-analysis/',
    license='LICENSE.txt',
    long_discription="Readme.md",
    author='GlazNik',
    author_email='golzitskij.ns@edu.spbstu.ru',
    description='David Cope\'s SPEAC-analysis in Python',
    python_requires='>=3.0'
)
